from GOESLayer import GOESLayer
import sys
import os
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/frameTypes'))
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/opsMsgs'))
from MPDUMsg import MPDUMsg
from CPPDUMsg import CPPDUMsg
from CPPDUOpsMsg import CPPDUOpsMsg
from CRC import CRC
import logging
import copy
from collections import deque

##
# @brief CPPDU Layer class.
# @details Fourth layer of LRIT packet processing. Extracts Transport frames
# from MPDU. Unlike previous layers, there is not generally a 1:1 relationship
# between input and output. This time, the relationship is many to many.
# This layer also juggles multiple Virtual Channel IDs (VCIDs) and APIDs.
# This is generally a port of the GOESProcessor implementation of the CPPDU layer
# logic, with some adjustments made in the interest of cleanliness and being
# "pythonic"
# @warning This is pretty messy, and there are definitely opportunities to
# simplify this logic. This component took so long to write and had so many
# problems that I ended up carefully copying and adapting the GOESProcessor
# code to get it working in the end
# @author Miguel Dominguez
# @date 2016
class CPPDULayerPort(GOESLayer):

   ##
   # @brief Transport State class.
   # @details These variables are common to every APID in progress.
   #The C++ has an awkward cascade of arrays and variables to manage
   #the state of the Transport Frame. Instead an object with multiple
   #instances seems more appropriate
   class TransportState(object):

      INVALID_APID = 2049

      ##
      # @fn __init__
      # @brief Constructor.
      # @details Initializes transport frame state to sane defaults.  If you
      # compare this with GOESProcessor.cpp you'll see fewer variables,
      # but that's because some of them are contained in CPPDUMsg
      def __init__(self):
         self.fullHeader = False #Whether we have the full header in hand
         self.extension = False #Whether the Transport Frame has data beyond what we have
         self.apidIndex = self.INVALID_APID #Pass to PPDU, this is the index of the file dict
         self.pendingApid = CPPDUMsg.APID_FILL_PACKET

   ##
   # @fn __init__
   # @brief Constructor.
   # @details Initializes the state variables that are not specific to a given
   # APID.  There may be some opportunity for cleanup here as I have several
   # dicts that all have keys based on APID
   def __init__(self):
       GOESLayer.__init__(self)
       self.activeApids = dict()
       self.nextState = dict()
       self.bytesRemaining = dict()
       self.output = dict()
       self.outOpsMsg = dict()
       self.currentVcid = 0
       self.logger = logging.getLogger(__name__)
       self.shippableProducts = deque()
       self.reset()
       self.framesWithLeftoverCount = 0
       self.outputCount = 0

   ##
   # @fn reset
   # @brief Resets state of current APID/VCID to sane defaults.
   # @details This function isn't in the C++ code but I found I need it
   # for some error conditions that the C++ does appear to address in this
   # layer
   def reset(self):
      self.currentState = self.TransportState()
      if self.currentVcid in self.nextState:
         self.nextState[self.currentVcid] = self.TransportState()
      #self.getApid = False #Whether we should seek a new header
      if self.currentState.apidIndex in self.output:
         self.output[self.currentState.apidIndex]  = CPPDUMsg()
      if self.currentState.apidIndex in self.outOpsMsg:
         self.outOpsMsg[self.currentState.apidIndex] = CPPDUOpsMsg()
      self.framePointer = 0
      self.remainingBytes = MPDUMsg.DATA_SIZE_BYTES
      self.packetHeader = []

   ##
   # @fn getLritPacketHeader
   # @brief Extracts a Transport header, if possible.  Updates state accordingly.
   # @details This is pretty much analogue to the C++ implementation, with a few
   # fixes.  There are opportunities for simplification here.  One thing that
   # annoys me are the multiple return statements, it might be nice to refactor
   # so that only one return is required at the end
   # @param[in] data The raw data bytes from the MPDU frame
   # @param[out] Success True if a packet was successfully extracted, false if
   # one was not for various reasons.  These include the MPDU frame not
   # containing the entire header, invalid header values found, and a new file
   # beginning in the middle of a sequence
   def getLritPacketHeader(self,data):
      if self.framePointer > MPDUMsg.DATA_SIZE_BYTES:
         self.logger.debug('Frame Pointer too large! {0}'.format(self.framePointer))
#         self.reset()
#         return False
      while len(self.packetHeader) < CPPDUMsg.HEADER_SIZE_BYTES:
         self.logger.debug('Frame Pointer: {0} PacketLen: {1}'.format(self.framePointer,len(self.packetHeader)))         
         self.packetHeader.append(data[self.framePointer])
         self.framePointer += 1
         self.remainingFrameBytes -= 1
         if self.remainingFrameBytes <= 0:
            self.nextState[self.currentVcid].fullHeader = False
            self.nextState[self.currentVcid].extension = True
            return False

#      self.logger.debug('New Header Found!')

      output = CPPDUMsg()
      #FIX THIS
      #self.logger.debug('Packet Header Length: {0}'.format(len(self.packetHeader)))
      output.getHeader(self.packetHeader)
      self.logger.debug(output)
      self.packetHeader = []
      if output.packetLength > CPPDUMsg.MAX_PACKET_LENGTH:
          self.logger.debug('packetlength too long!')          
          self.reset()
          return False
      self.output[output.apid] = output
      self.outOpsMsg[output.apid] = CPPDUOpsMsg()
      self.nextState[self.currentVcid].fullHeader = True
      
      if output.apid == CPPDUMsg.APID_FILL_PACKET:
         self.currentState.apidIndex = output.apid
         return True
      
      #If this is already an active APID, keep on going
      self.logger.debug(self.activeApids)
      if output.apid in self.activeApids:
         self.logger.debug('Old Output Apid')
         #self.logger.debug(output)
         self.currentState.apidIndex = output.apid
         self.bytesRemaining[self.currentState.apidIndex] = output.packetLength
         self.nextState[self.currentVcid].apidIndex = self.currentState.apidIndex
         return True

      #If this is not an active APID, and the sequence flag does not suggest that
      #this is the start of a new APID, then fail
      if (output.sequenceFlag != 3) and (output.sequenceFlag != 1):
         self.logger.debug('Unknown New File')
         return False

      #Otherwise, add new active APID and keep on going
      self.logger.debug('New Output Apid')
      self.activeApids[output.apid] = True
      self.logger.debug(self.activeApids)
      self.currentState.apidIndex = output.apid
      self.bytesRemaining[self.currentState.apidIndex] = output.packetLength
      self.nextState[self.currentVcid].apidIndex = self.currentState.apidIndex
      
      return True

   ##
   # @fn attemptCompletePacket
   # @brief Checks if packet is full-size, and if so performs some finishing operations
   # @details This function calculates CRC, adds the outcome of the CRC to an Ops Msg,
   # and pushes the complete transport frame to a queue.  Why does it push to a queue?
   # Because a single MPDU Frame can result in multiple complete Transport frames
   # theoretically, so we want to take this possibility into account
   def attemptCompletePacket(self):
      if self.output[self.currentState.apidIndex].packetLength == len(self.output[self.currentState.apidIndex].data):
 #        self.logger.debug('Full packet!')
         #Complete output
         self.output[self.currentState.apidIndex].getCrcFromData()
         hasErrors = not CRC.Calculate(self.output[self.currentState.apidIndex].data,self.output[self.currentState.apidIndex].crc)

         self.outOpsMsg[self.currentState.apidIndex].setHasCrcErrors(hasErrors)
         #Make note of CRC errors if they exist
         if self.outOpsMsg[self.currentState.apidIndex].getHasCrcErrors():
             self.logger.debug('CRC Errors Found!')
        #     self.reset()
         #Remove APID from current list of APIDs
         self.shippableProducts.appendleft((self.output[self.currentState.apidIndex],self.outOpsMsg[self.currentState.apidIndex]))

   ##
   # @fn consumeData
   # @brief Processes an MPDU frame and updates state accordingly.
   # @details This is pretty much analogue to the C++ implementation, with a few
   # fixes.  There are opportunities for simplification here.  One thing that
   # annoys me are the multiple return statements, it might be nice to refactor
   # so that only one return is required at the end
   # @param[in] data The MPDUMsg object
   def consumeData(self,data):

      self.getApid = False
     # self.logger.debug('FHP: {0}'.format(data.firstHeaderPointer))
      self.currentState = copy.deepcopy(self.nextState[self.currentVcid])
      
      self.framePointer = 0
      #self.logger.debug('Frame Pointer Reset')
      self.remainingFrameBytes = MPDUMsg.DATA_SIZE_BYTES

      if data.firstHeaderPointer > MPDUMsg.DATA_SIZE_BYTES and data.firstHeaderPointer < MPDUMsg.NOT_NEW_FRAME:
         self.logger.debug('Invalid FHP: {0}'.format(data.firstHeaderPointer))
         #self.reset()
         #return


      if data.firstHeaderPointer == MPDUMsg.NOT_NEW_FRAME:

         if self.currentState.pendingApid == CPPDUMsg.APID_FILL_PACKET:

            self.nextState[self.currentVcid].fullHeader = False 
            self.nextState[self.currentVcid].extension = False 
            self.nextState[self.currentVcid].apidIndex = self.TransportState.INVALID_APID
            self.nextState[self.currentVcid].pendingApid = CPPDUMsg.APID_FILL_PACKET
            return
         if (not self.currentState.fullHeader) and self.currentState.extension:

            #Theoretically at this point that you will have enough data to get the packet header
            self.getLritPacketHeader(data.data)
            self.logger.debug('New APID: {0}'.format(self.output[self.currentState.apidIndex].apid))
            if self.output[self.currentState.apidIndex].apid == CPPDUMsg.APID_FILL_PACKET:

               if self.output[self.currentState.apidIndex].packetLength < self.remainingFrameBytes:
                  self.framePointer += self.output[self.currentState.apidIndex].packetLength
                  self.remainingFrameBytes -= self.output[self.currentState.apidIndex].packetLength
                  self.logger.debug('Packet Length added to Frame Pointer (FHP 2047): {0} remainingFrameBytes: {1}'.format(self.framePointer,self.remainingFrameBytes))
                  self.getApid = True

               elif self.output[self.currentState.apidIndex].packetLength == self.remainingFrameBytes:
                  self.nextState[self.currentVcid].extension = False
                  return

               if not self.getApid:
                  self.nextState[self.currentVcid].extension = True
                  self.nextState[self.currentVcid].pendingApid = CPPDUMsg.APID_FILL_PACKET
                  self.bytesRemaining[CPPDUMsg.APID_FILL_PACKET] = self.output[self.currentState.apidIndex].packetLength - self.remainingFrameBytes
                  return
            
            if not self.getApid:
               self.currentState.pendingApid = self.nextState[self.currentVcid].apidIndex
         if not self.getApid:

            if not self.currentState.extension:
               return

            self.currentState.apidIndex = self.currentState.pendingApid

            if self.bytesRemaining[self.currentState.apidIndex] < (MPDUMsg.DATA_SIZE_BYTES - self.framePointer):
               #self.logger.debug('{0}'.format(self.bytesRemaining))
               return

            self.output[self.currentState.apidIndex].fromFrame(data.data[self.framePointer:(MPDUMsg.DATA_SIZE_BYTES)])
            self.attemptCompletePacket()
            self.bytesRemaining[self.currentState.apidIndex] -= (MPDUMsg.DATA_SIZE_BYTES - self.framePointer)
            return

      if (data.firstHeaderPointer > 0) or not self.getApid:
         
         if not self.currentState.fullHeader and self.currentState.extension:
            
            self.getLritPacketHeader(data.data)
            self.currentState.pendingApid = self.nextState[self.currentVcid].apidIndex

         if self.currentState.extension:
            self.currentState.apidIndex = self.currentState.pendingApid

            self.output[self.currentState.apidIndex].fromFrame(data.data[self.framePointer:(self.framePointer + self.bytesRemaining[self.currentState.apidIndex])])
            self.attemptCompletePacket()
            self.remainingFrameBytes -= self.bytesRemaining[self.currentState.apidIndex]
            self.framePointer += self.bytesRemaining[self.currentState.apidIndex]
            self.logger.debug('Bytes Remaining added to Frame Pointer: {0} bytesRemaining: {1}'.format(self.framePointer,self.bytesRemaining))

            if self.output[self.currentState.apidIndex].sequenceFlag == 3 or self.output[self.currentState.apidIndex].sequenceFlag == 2: 
               self.logger.debug('Popping Active APID')
               self.activeApids.pop(self.output[self.currentState.apidIndex].apid,None)
         else:
            self.framePointer += data.firstHeaderPointer
            self.logger.debug('First Header Pointer added to Frame Pointer: {0} FHP: {1}'.format(self.framePointer,data.firstHeaderPointer))
            self.remainingFrameBytes -= data.firstHeaderPointer

         self.getApid = True

      if (data.firstHeaderPointer == 0) or self.getApid:
         self.getApid = True
         bytesToWrite = 0
         while self.getApid:
            if not self.getLritPacketHeader(data.data):
               return
            if self.output[self.currentState.apidIndex].apid == CPPDUMsg.APID_FILL_PACKET:
               if self.output[self.currentState.apidIndex].packetLength < self.remainingFrameBytes:
                  self.framePointer += self.output[self.currentState.apidIndex].packetLength
                  self.remainingFrameBytes -= self.output[self.currentState.apidIndex].packetLength
                  self.logger.debug('Packet Length added to Frame Pointer (FHP 0): {0} remainingFrameBytes: {1}'.format(self.framePointer,self.remainingFrameBytes))
                  self.getApid = True
               else:
                  if self.output[self.currentState.apidIndex].packetLength == self.remainingFrameBytes:
                     self.nextState[self.currentVcid].extension = False
                     return
                  self.nextState[self.currentVcid].extension = True
                  self.nextState[self.currentVcid].pendingApid = CPPDUMsg.APID_FILL_PACKET
                  self.bytesRemaining[CPPDUMsg.APID_FILL_PACKET] = self.output[self.currentState.apidIndex].packetLength - self.remainingFrameBytes
            else:
               self.getApid = False

            if not self.getApid:
               if self.bytesRemaining[self.currentState.apidIndex] <= self.remainingFrameBytes:
                  bytesToWrite = self.bytesRemaining[self.currentState.apidIndex]
                  self.nextState[self.currentVcid].extension = False
               else:
                  bytesToWrite = self.remainingFrameBytes
                  self.bytesRemaining[self.currentState.apidIndex] -= self.remainingFrameBytes
                  self.currentState.pendingApid = self.nextState[self.currentVcid].apidIndex
                  self.nextState[self.currentVcid].pendingApid = self.currentState.pendingApid
                  self.nextState[self.currentVcid].extension = True

               self.output[self.currentState.apidIndex].fromFrame(data.data[self.framePointer:(self.framePointer + bytesToWrite)])
               self.attemptCompletePacket()
               self.remainingFrameBytes -= bytesToWrite;
               self.framePointer += bytesToWrite;
               
               if self.nextState[self.currentVcid].extension:
                  return

               if self.output[self.currentState.apidIndex].sequenceFlag == 3 or self.output[self.currentState.apidIndex].sequenceFlag == 2: 
                  self.logger.debug('Popping Active APID')
                  self.activeApids.pop(self.output[self.currentState.apidIndex].apid,None)

               #if data buffer is empty
               if self.remainingFrameBytes <= 0:
                  return

               self.getApid = True


   ##
   # @fn threadFunction
   # @brief Custom thread function that allows multiple outputs per run.
   def threadFunction(self):
         while True:
            (inData,inOpsMsg) = self.inBuffer.get()
            if inData is not None:
               self.processData(inData,inOpsMsg)
               while self.shippableProducts:
                  self.outBuffer.put(self.shippableProducts.pop())
                  self.outputCount += 1
                  self.logger.info('Output Count: {0}'.format(self.outputCount))
            else:
               self.outBuffer.put((None,None))
               break

   ##
   # @fn processData
   # @brief Extracts Transport data from MPDU and performs some basic checks.
   # @detail Checks if there are missed frames and resets state if there are.
   # Then the MPDU data is processed.
   # @param[in] data MPDUMsg Object
   # @param[in] opsMsgs Information about missed frames
   # @param[out] outData CPPDUMsg object
   # @param[out] outOpsMsg CPPDUOpsMsg Object, mentioning CRC failures if they
   # occurred
   def processData(self,data,opsMsg):
      self.currentVcid = opsMsg.vcID
      if self.currentVcid not in self.nextState:
         self.nextState[self.currentVcid] = self.TransportState()
      if opsMsg.hasMissedFrames():
         self.logger.debug('Resetting!')
         self.reset()
      self.consumeData(data)
