##
# @brief VCDU Operations Message class.
# @author Miguel Dominguez
# @details There is no superclass for Operations Messages Classes because
# they basically have nothing in common, they are simply a container for
# some useful variables to be passed forward.
# @date 2016
class VCDUOpsMsg(object):

   ##
   # @fn __init__
   # @brief Constructor.
   # @details Initializes the message states to sane defaults
   def __init__(self):
      self.rsPass = [True, True, True, True]
      self.isFill = False

   ##
   # @fn RSFail
   # @brief Report that one of the Reed-Solomon channels failed to decode
   # @param[in] index The index of the particular RS channel that failed  
   # @warning This may not be strictly necessary, as the VCDULayer does not
   # forward data with even one failure in any channel
   def RSFail(self,index):
      self.rsPass[index] = False

   ##
   # @fn RSNoErrors
   # @brief Return whether all of the Reed-Solomon channels were reported
   # to have succeeded
   # @param[out] output True if all RS channels passed, False if even one
   # of them failed
   # @warning This may not be strictly necessary, as the VCDULayer does not
   # forward data with even one failure in any channel
   def RSNoErrors(self):
      return len([x for x in self.rsPass if x == False]) == 0

   ##
   # @fn RSErrorLocations
   # @brief Return a list of indices of channels where Reed-Solomon
   # decoding failed
   # @param[out] output List of indices of failed channels. This is
   # an empty list if there were no failures.
   # @warning This may not be strictly necessary, as the VCDULayer does not
   # forward data with even one failure in any channel
   def RSErrorLocations(self):
      return [i for i in range(len(self.rsPass)) if self.rsPass[i] == False] or [None]

   ##
   # @fn setIsFill
   # @brief Report whether a VCDU packet is marked as "Fill"
   # @param[in] isFill whether a VCDU packet is marked as "Fill"
   # @warning This may not be strictly necessary, as the VCDULayer is now
   # smart enough not to forward fill packets.
   def setIsFill(self,isFill):
      self.isFill = isFill
