#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 The Aerospace Corporation.
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
#

import numpy as np
from gnuradio import gr
import threading as th
import pmt
import time

from collections import deque, namedtuple

class tagged_stream_frame_sync(gr.basic_block):
    """
    Output frames synchronized to tagged input stream.
    Look for consecutive tags separated by the expected number of samples, with
    optional tolerance.  Once found, declare lock and pass fixed length frames to
    output.  Unlock after some number of consecutive missed tags.
    """
    def __init__(self, itemsize, tagstring, frame_len, lock_frames, unlock_frames, msgport=True, tol=1, timeout=None):
        """
        @param itemsize bytes per input item
        @param tagstring comma separated strings for the tag keys that indicates the first item in frame (max 6)
        @param frame_len number of items in a frame
        @param lock_frames lock after this many consecutive detected frame tags; set to 0 to lock on first tag
        @param unlock_frames unlock after missing this many consecutive frame tags
        @param msgport send messages on state change (default:True)
        @param tol frame tag detected if tag in frame_len +/- tol
        @param timeout unlock if no input items in timeout seconds; 0 for no timeout

        """

        assert frame_len > 0
        assert lock_frames > 0
        assert unlock_frames > 0


        gr.basic_block.__init__(self, name="tagged_stream_frame_sync",
                                in_sig=[(np.byte, itemsize)],
                                out_sig=[(np.byte, itemsize)])

        self.locked = th.Event()                # enable external triggers in other threads
        self.lost_lock = th.Event()             # True when unlocked since last poll
        tagkeys = [pmt.intern(s.strip()) for s in tagstring.split(',')]
        self._tag_keys = getattr(pmt, 'list%d' % min(len(tagkeys), 6))(*tagkeys)
        self.frame_size = frame_len
        self.lock_frames = lock_frames
        self.unlock_frames = self._to_unlock = unlock_frames

        self._timeout = timeout                 # inactivity timeout
        self.last_reset_time = 0                # last time general work was called

        # custom named tuple: sample offset of tag, frames left until lock
        self.LockTuple = namedtuple('LockTuple', ['offset', 'to_lock'])
        self._lock_tracks = deque()             # contains candidate LockTuples
        self._last_lock = 0                  # contains last locked offset

        self.set_tag_propagation_policy(gr.TPP_DONT)  # Must account for dropped samples
        self.intol = set(range(self.frame_size - tol,
                               self.frame_size + tol + 1))  # tag interval range

        # local loggers
        name = self.to_basic_block().alias()
        loggers = [x for x in gr.logger_get_names() if name in x]
        gr.log.log_assert(len(loggers) == 2, "Too many loggers for %s" % name)
        self.log, self.log_debug = tuple(gr.logger(n) for n in sorted(loggers))

        self.running = th.Event()               # for inactivity thread
        self.inactive_thread = None

        # register message port to notify state changes
        self.msgport = pmt.intern('state') if msgport else None
        if self.msgport:
            self.message_port_register_out(self.msgport)

    def forecast(self, noutput_items, ninput_items_required):
        #setup size of input_items[i] for work call
        for i, _ in enumerate(ninput_items_required):
            ninput_items_required[i] = noutput_items

    def start(self):
        if self._timeout:
            self.inactive_thread = th.Thread(target=self.inactivity_timeout,
                                             name='inactivity_thread')
            self.inactive_thread.start()
            self.last_reset_time = time.time()
            self.running.set()
        return True

    def stop(self):
        if self.running.is_set():
            self.running.clear()
            self.inactive_thread.join()
        return True

    def general_work(self, input_items, output_items):
        noutput_items = output_items[0].size

        self.last_reset_time = time.time()

        # get all tags to avoid retrieving again in _propagate_tags
        tags = self.get_tags_in_range(0, self.nitems_read(0),
                                      self.nitems_read(0) + noutput_items)

        # get tags with sync key
        sync_tags = deque(t for t in tags if pmt.list_has(self._tag_keys, t.key))

        self.log_debug.debug("noutput_items = %d, nitems_read(0) = %d, sync_tags @ %s" %
                             (noutput_items, self.nitems_read(0),
                              ', '.join('%d' % x.offset for x in sync_tags)))

        # TODO: this does not handle the case where one stream is unlocking
        # while a new stream is starting to lock.  This and the locked section
        # could prob. be merged and overlapping lock/unlock processes could be
        # handled.
        if not self.locked.is_set():
            # have we locked?

            # TODO: this is actually a really tough problem.  If you get a lot of
            # tags you may have multiple possible starting points and multiple
            # parallel tag sets to track until you find one that locks.
            for tag in sync_tags:
                next_tuple = None
                to_lock = self.lock_frames - 1   # default frames until lock (-1 for current tag)
                while self._lock_tracks:
                    next_tuple = self._lock_tracks.popleft()
                    # if the next tag is beyond the current first_lock tag
                    # then the first_lock tag chain is broken, remove
                    if tag.offset - next_tuple.offset > max(self.intol):
                        self.log_debug.info('Remove tag @ %d / %d' % next_tuple)
                        continue

                    # if in tolerance, update frames until lock
                    if tag.offset - next_tuple.offset in self.intol:
                        self.log_debug.info('Remove to update tag @ %d / %d' % next_tuple)
                        to_lock = next_tuple.to_lock - 1
                    break

                if to_lock == 0:   # LOCK!
                    #Add tag to help compute missed VCDUs in protoproc block
                    self.add_item_tag(0, tag.offset,pmt.to_pmt("lock"), pmt.to_pmt(tag.offset-self._last_lock))
                    self._lock_sync('Locked @ %d, nitems_written = %d',
                                    tag.offset, self.nitems_written(0))
                    noutput_items = tag.offset - self.nitems_read(0)
                    break                               # consume up to state change

                # update prior tuple or make new one w/ the latest offset/to_lock
                try:
                    next_tuple = next_tuple._replace(offset=tag.offset, to_lock=to_lock)
                except AttributeError:   # next_tuple still None b/c self._lock_tracks was empty
                    next_tuple = self.LockTuple(offset=tag.offset, to_lock=to_lock)

                self.log_debug.info('Add tag @ %d / %d' % next_tuple)
                self._lock_tracks.append(next_tuple)

            # consume up to lock point but don't output anything
            #self.log_debug.debug("Unlocked: Consume %d" % noutput_items)
            self.consume(0, noutput_items)
            return 0

        ## Below this point we are locked ##
        assert self.locked.is_set()

        # have we maintained lock?
        while sync_tags:
            tag = sync_tags.popleft()
            delta = tag.offset - self._last_lock
            # this assumes tags are in chronological order
            if delta in self.intol:    # hit! Still in lock
                self.log_debug.info('Hit! Maintain lock from tag @ %d, delta = %d' %
                                    (tag.offset, delta))
                self._last_lock = tag.offset                    # update last lock
                self._to_unlock = self.unlock_frames            # reset counter
                continue                                        # go to next tag

            # if this tag is in a subsequent frame
            if delta > max(self.intol):
                self._to_unlock -= 1                # missed tag in this frame
                self._last_lock += self.frame_size  # go to the next frame
                self.log_debug.info('Miss! Tag @ %d in next frame, last_lock => %d, to_unlock => %d' %
                                    (tag.offset, self._last_lock, self._to_unlock))
                sync_tags.appendleft(tag)            # keep this tag for next iteration

                if self._to_unlock == 0:                # UNLOCK!
                    # consume / produce up to state change (self._last_lock)
                    noutput_items = self._last_lock - self.nitems_read(0)
                    self._unlock_sync('Unlock! Next tag @ %d, last_lock = %d ',
                                      tag.offset, self._last_lock)
                    break

        else:   # all tags processed without unlock
            # Here, self._last_lock has latest lock point at least at nitems_read(0)
            # and self._to_unlock has frames until unlock event.

            # check if unlock in the remainder of the frame
            items_after_lock = noutput_items - self._last_lock + self.nitems_read(0)
            self.log_debug.debug('Items after lock = %d, nitems_read = %d, _last_lock = %d' %
                                 (items_after_lock, self.nitems_read(0), self._last_lock))

            frames_after_lock, rem_after_lock = divmod(items_after_lock, self.frame_size)

            # make sure we don't end up inside the current tolerance window
            if rem_after_lock + self.frame_size in self.intol:
                frames_after_lock -= 1
                rem_after_lock += self.frame_size

            # make sure we don't end up within next tolerance window
            rem_after_lock = min(rem_after_lock, min(self.intol))

            # don't go past unlock event
            frames_after_lock = min(frames_after_lock, self._to_unlock)
            self._last_lock += frames_after_lock * self.frame_size
            self._to_unlock -= frames_after_lock

            if self._to_unlock == 0:    # UNLOCK!
                rem_after_lock = 0      # don't go past state change (unlock)
                self._unlock_sync('Unlock! No tags in [%d, %d)',
                                  self._last_lock - frames_after_lock * self.frame_size,
                                  self._last_lock)
            elif frames_after_lock > 0:
                self.log_debug.info('Miss! No tags, maintain lock: '
                                    'last_lock => %d, to_unlock => %d' %
                                    (self._last_lock, self._to_unlock))
                self.log_debug.debug('frames_after_lock = %d, rem_after_lock = %d' %
                                     (frames_after_lock, rem_after_lock))

            noutput_items = self._last_lock - self.nitems_read(0) + rem_after_lock

        # propagate and consume up to noutput_items
        self.log_debug.debug("Locked: Consume & produce %d" % noutput_items)
        self._propagate_tags([t for t in tags if t.offset < noutput_items + self.nitems_read(0)])
        output_items[0][:noutput_items] = input_items[0][:noutput_items]
        self.consume(0, noutput_items)
        return noutput_items

    def _propagate_tags(self, tags, which_output=0):
        # Propagate tags
        samples_dropped = self.nitems_read(which_output) - self.nitems_written(which_output)
        for tag in tags:
            tag.offset -= samples_dropped
            self.add_item_tag(which_output, tag)

    def _lock_sync(self, logmsg=None, *logargs):
        """Change to the locked state."""
        logmsg and self.log_debug.info(logmsg % logargs)
        self._last_lock = logargs[0]    # tag.offset
        self._to_unlock = self.unlock_frames
        self.locked.set()
        if self.msgport:
            self.message_port_pub(self.msgport, pmt.cons(pmt.intern('lock'), pmt.PMT_T))

    def _unlock_sync(self, logmsg=None, *logargs):
        """Change to the unlocked state."""
        logmsg and self.log_debug.info(logmsg % logargs)
        self.lost_lock.set()
        self._lock_tracks.clear()            # reset container
        self.locked.clear()
        if self.msgport:
            self.message_port_pub(self.msgport, pmt.cons(pmt.intern('lock'), pmt.PMT_F))

    # TODO: if you don't do anything but return a member, why make this function at all?
    def get_locked(self):
        """
        Are we currently locked
        """
        return self.locked.is_set()

    def get_lost_lock(self):
        """
        Have we lost lock since last check (even if regained since)
        """
        result = self.lost_lock.is_set()
        self.lost_lock.clear()
        return result

    # TODO: this is a stand-in for real MAC code running outside this PHY block
    def inactivity_timeout(self):
        running = self.running
        running.wait()
        while running.is_set():
            time.sleep(self._timeout)   # TODO: make this shorter to exit more quickly?
            time_delta = time.time() - self.last_reset_time
            if time_delta > self._timeout:
                if self.locked.is_set():
                    self.log_debug.info('Unlock! Inactivity timeout: %.1f' % time_delta)
                    self.locked.clear()
                    self.lost_lock.set()
            #self.log_debug.debug('Reset inactivity timeout')

######################## TEST CODE #################################

from gnuradio import blocks
from gnuradio.gr import tag_utils
import pmt

import time

def test1():
    # test using message to pdu stream

    tb = gr.top_block()
    # Possibly use the message_strobe(_random) blocks to send messages into a
    # PDU to tagged stream block that will output streams with tags and then
    # input into this block to see how it behaves

    offset = 100
    frame_size = 100
    gap_size = 88
    tag_key = 'TAG!'
    lock_after, unlock_after = 5, 4

    # dictionary {k, v} defines tags with key k and value v
    msg_meta = pmt.make_dict()
    msg_meta = pmt.dict_add(msg_meta, pmt.intern(tag_key), pmt.intern('acq tag'))

    data = [x + 0j for x in range(frame_size)]
    dmsg = pmt.cons(msg_meta, pmt.init_c32vector(frame_size, data))
    gmsg = pmt.cons(pmt.PMT_NIL, pmt.init_c32vector(gap_size, [-1 + 0j,] * gap_size))

    rate = 110
    period = frame_size / rate * 1e3   # ms for msg strobe, samples for tagstrobe

    src = blocks.message_strobe(dmsg, period)
    pdu2stream = blocks.pdu_to_tagged_stream(blocks.complex_t)
    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)

    fsync = tagged_stream_frame_sync(100, tag_key, lock_after, unlock_after)
    snk = blocks.null_sink(gr.sizeof_gr_complex)

    tb.msg_connect(src, 'strobe', pdu2stream, 'pdus')
    tb.connect(pdu2stream, throttle, fsync, snk)
    try:
        tb.start()
        print "------------------> Start tags"
        while not fsync.get_locked():
            time.sleep(1)
        time.sleep(2)
        print "------------------> Stop tags"
        src.to_basic_block()._post(pmt.intern('set_msg'), gmsg)
        while fsync.get_locked():
            time.sleep(1)
    except KeyboardInterrupt:
        pass
    finally:
        src.to_basic_block()._post(pmt.intern('system'), pmt.cons(pmt.intern("done"), pmt.PMT_T))
        tb.stop()
        tb.wait()

def test2():
    # test using vector source

    tb = gr.top_block()
    # Possibly use the message_strobe(_random) blocks to send messages into a
    # PDU to tagged stream block that will output streams with tags and then
    # input into this block to see how it behaves

    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (10, 25, 310, 400, 2, 6)
    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 100, 310, 210, 4, 5)
    offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 100, 610, 210, 4, 5)
    tag_key = 'TAG!'
    rate = 100e3
    num_tags = 10
    data_size = (num_tags + gap_size // frame_size + 1) * frame_size

    data = [x + 0j for x in range(data_size)]
    tagoffsets = [(offset + frame_size * x) % data_size for x in range(num_tags)]
    tagoffsets = [x if x < gap_offset else x + gap_size for x in tagoffsets]
    taglist = [tag_utils.python_to_tag((x, pmt.intern(tag_key), pmt.PMT_T, pmt.PMT_NIL))
               for x in tagoffsets]
    src = blocks.vector_source_c(data, repeat=True, tags=taglist)
    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)

    fsync = tagged_stream_frame_sync(frame_size, tag_key, lock_after, unlock_after)
    snk = blocks.null_sink(gr.sizeof_gr_complex)

    print "--->> data_size = ", data_size, ", tag offsets @ ", sorted(tagoffsets)
    print "--->> frame_size = ", frame_size, ", unlock_after = ", unlock_after, ", lock_after = ", lock_after
    tb.connect(src, throttle, fsync, snk)
    try:
        tb.start()
        time.sleep(10)
    except KeyboardInterrupt:
        pass
    finally:
        print "STOP"
        tb.stop()
        tb.wait()

def test3():
    # test using tag strobe

    tb = gr.top_block()

    offset = 22
    frame_size = 100
    gap_size = 343
    tag_key = 'TAG!'
    lock_after, unlock_after = 5, 4
    rate = 100e3

    src = blocks.tags_strobe(gr.sizeof_gr_complex, pmt.to_pmt(1), frame_size, pmt.intern(tag_key))
    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)
#    tagdbg = blocks.tag_debug(gr.sizeof_gr_complex, "tag debug")

    fsync = tagged_stream_frame_sync(100, tag_key, lock_after, unlock_after)
    snk = blocks.null_sink(gr.sizeof_gr_complex)

    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)

    fsync = tagged_stream_frame_sync(frame_size, tag_key, lock_after, unlock_after)
    snk = blocks.null_sink(gr.sizeof_gr_complex)

    tb.connect(src, throttle, fsync, snk)
    try:
        tb.start()
        time.sleep(10)
    except KeyboardInterrupt:
        pass
    finally:
        print "STOP"
        tb.stop()
        tb.wait()


class test_valve_c(gr.basic_block):
    def __init__(self, isopen=False):
        gr.basic_block.__init__(self, "test_valve", [np.complex64], [np.complex64])
        self.isopen = isopen

    def general_work(self, input_items, output_items):
        if self.isopen:
            self.consume(0, len(input_items[0]))
            return 0
        noutput_items = len(output_items[0])
        output_items[0] = input_items[:noutput_items]
        self.consume(0, noutput_items)
        return noutput_items

def test4():
    # test using vector source and test valve to test timer timeout
    tb = gr.top_block()

    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (10, 25, 310, 400, 2, 6)
    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 100, 310, 210, 4, 5)
    offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 100, 610, 210, 4, 5)
    tag_key = 'TAG!'
    rate = 100e3
    num_tags = 10
    data_size = (num_tags + gap_size // frame_size + 1) * frame_size

    data = [x + 0j for x in range(data_size)]
    tagoffsets = [(offset + frame_size * x) % data_size for x in range(num_tags)]
    tagoffsets = [x if x < gap_offset else x + gap_size for x in tagoffsets]
    taglist = [tag_utils.python_to_tag((x, pmt.intern(tag_key), pmt.PMT_T, pmt.PMT_NIL))
               for x in tagoffsets]
    src = blocks.vector_source_c(data, repeat=True, tags=taglist)
    valve = test_valve_c()
    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)

    fsync = tagged_stream_frame_sync(frame_size, tag_key, lock_after, unlock_after)
    snk = blocks.null_sink(gr.sizeof_gr_complex)

    print "--->> data_size = ", data_size, ", tag offsets @ ", sorted(tagoffsets)
    print "--->> frame_size = ", frame_size, ", unlock_after = ", unlock_after, ", lock_after = ", lock_after
    tb.connect(src, throttle, valve, fsync, snk)
    try:
        tb.start()
        time.sleep(1)
        while not fsync.locked:
            time.sleep(1)
        print "--->> Open valve"
        valve.isopen = True
        print "--->> Open valve"
        while fsync.locked:
            time.sleep(1)
    except KeyboardInterrupt:
        pass
    finally:
        print "STOP"
        tb.stop()
        tb.wait()

def test5():
    # test using ramp to test passed data
    tb = gr.top_block()

    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (10, 25, 310, 400, 2, 6)
    #offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 100, 310, 210, 4, 5)
    offset, frame_size, gap_size, gap_offset, lock_after, unlock_after = (22, 10, 610, 210, 4, 5)
    tag_key = 'TAG!'
    rate = 100e3
    num_tags = 10
    data_size = (num_tags + lock_after + unlock_after) * frame_size

    # data < 0 b/4 lock, > 0 after lock, should only see >= 0 at output
    data = [x - (lock_after - 1) * frame_size + 0j for x in range(data_size)] + [0,] * gap_size
    tagoffsets = [(frame_size * x) % data_size for x in range(num_tags)]
    tagoffsets = [x if x < gap_offset else x + gap_size for x in tagoffsets]
    taglist = [tag_utils.python_to_tag((x, pmt.intern(tag_key), pmt.PMT_T, pmt.PMT_NIL))
               for x in tagoffsets]
    src = blocks.vector_source_c(data, repeat=True, tags=taglist)
    dly = blocks.delay(gr.sizeof_gr_complex, offset)                # offset samples rather than tag
    throttle = blocks.throttle(gr.sizeof_gr_complex, rate)

    fsync = tagged_stream_frame_sync(gr.sizeof_gr_complex, tag_key, frame_size, lock_after, unlock_after)
    snk = blocks.vector_sink_c()

    print "--->> data_size = ", data_size, ", tag offsets @ ", sorted(tagoffsets)
    print "--->> frame_size = ", frame_size, ", unlock_after = ", unlock_after, ", lock_after = ", lock_after
    tb.connect(src, throttle, dly, fsync, snk)
    try:
        tb.start()
        time.sleep(2)
        print "--> wait for lock"
        fsync.locked.wait()
        print "--> wait for unlock"
        while fsync.locked.is_set():
            time.sleep(0)
        fsync.locked.wait()
    except KeyboardInterrupt:
        pass
    finally:
        print "STOP"
        tb.stop()
        tb.wait()

    import itertools as itt
    rxdata = snk.data()
    start_offset = (lock_after - 1) * frame_size
    unlock_size = (unlock_after + num_tags - lock_after) * frame_size
    print "PASSED %d samples" % (len(rxdata)) \
          if not any(x - y for x, y in itt.izip(itt.cycle(
                    data[start_offset:start_offset + unlock_size]), rxdata)) \
          else \
          "NOT PASSED %d" % (len(rxdata))

if __name__ == '__main__':
    test5()
