#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

import struct
from collections import namedtuple

## Main reference:
# LRIT Receiver Specification http://noaasis.noaa.gov/LRIT/system_specs.html

################################################################################
#                                                                              #
# Header section                                                               #
#                                                                              #
################################################################################

# header type codes
(COMMON_HTYPE,
 UNKNOWN_HTYPE, 
 PRIMARY_HTYPE,     # here and below are types defined in LRIT spec
 IMAGE_STR_HTYPE, 
 IMAGE_NAV_HTYPE, 
 IMAGE_DAT_HTYPE, 
 ANNOTATION_HTYPE, 
 TIMESTAMP_HTYPE, 
 TEXT_HTYPE,        # seems mislabeled as 7 in the spec, should be 6
 KEY_HTYPE,
 SEG_ID_HTYPE, 
# ENC_KEY_HTYPE,    # from TX spec, doesn't seem to be used now
 NOAA_LRIT_HTYPE,
 HEADER_STR_HTYPE, 
 RICE_HTYPE,
 FILENAME_HTYPE) = range(-2, 8) + [128, 129, 130, 131, 132]

# Header specification type
_HeaderTuple = namedtuple('_HeaderTuple', 'name, fmt, fields')

# Header spec map
_HEADER_SPECS = {
#   TYPE constant                Name       Parse string(1,2) Field names
    COMMON_HTYPE:   _HeaderTuple('Common',     '!BH',       'hdr_type, hdr_len'),
    UNKNOWN_HTYPE:  _HeaderTuple('Unknown',    '!BH%ds',    'hdr_type, hdr_len, data'),
    PRIMARY_HTYPE:  _HeaderTuple('Primary',    '!BHBIQ',    'hdr_type, hdr_len, file_type, tot_hdr_len, data_len'),
    IMAGE_STR_HTYPE:_HeaderTuple('ImageStr',   '!BHBHHB',   'hdr_type, hdr_len, bits_per_pxl, ncols, nlines, iscompressed'),
    IMAGE_NAV_HTYPE:_HeaderTuple('ImageNav',   '!BH32sIIII','hdr_type, hdr_len, proj, cfac, lfac, coff, loff'), 
    IMAGE_DAT_HTYPE:_HeaderTuple('ImageDat',   '!BH%ds',    'hdr_type, hdr_len, data'),
    ANNOTATION_HTYPE:_HeaderTuple('Annotation','!BH%ds',    'hdr_type, hdr_len, text'),
    TIMESTAMP_HTYPE:_HeaderTuple('TimeStamp',  '!BH7s',     'hdr_type, hdr_len, timestamp'),
    TEXT_HTYPE:     _HeaderTuple('Text',       '!BH%ds',    'hdr_type, hdr_len, text'),
    KEY_HTYPE:      _HeaderTuple('KeyStr',     '!BH%ds',    'hdr_type, hdr_len, key'),
    #SEG_ID_HTYPE:   _HeaderTuple('SegmentId',  '!BHBBHBB',  'hdr_type, hdr_len, seq_cnt, tot_seg, reserved, img_type, header_type'),
    SEG_ID_HTYPE:   _HeaderTuple('SegmentId',  '!BH7H',     'hdr_type, hdr_len, id, num_seg, col0, line0, tot_seg, max_col, max_row'),
                     # NOTE: Former SEGMENT ID HEADER from TX Spec is WRONG!!, use the one from RX SPEC
                     
    #ENC_KEY_HTYPE:  _HeaderTuple('EncKey',     '!BHH',      'hdr_type, hdr_len, auth_user'),
    NOAA_LRIT_HTYPE:_HeaderTuple('NoaaLrit',   '!BH4sHHHB', 'hdr_type, hdr_len, sig, pid, psid, param, iscompressed'),
                    # NOTE: Type 129 is NOT Encryption Key Header from TX spec, see RX Spec 
    HEADER_STR_HTYPE:_HeaderTuple('HeaderStr', '!BH%ds',    'hdr_type, hdr_len, data'),
    RICE_HTYPE:     _HeaderTuple('Rice',       '!BHHBB',    'hdr_type, hdr_len, flags, pxl_per_blk, nlines'),
    FILENAME_HTYPE: _HeaderTuple('Filename',   '!BH%ds',    'hdr_type, hdr_len, text')

}              
# NOTE: parse strings are according to struct module format strings
# (1): ! in parse string required for network byte alignment
# (2): % in parse string indicates variable field length

# experiment a little with metaclasses, allows dynamic class definitions based
# on a single class attribute (_type)
class _MetaLritHeader(type):
    def __new__(cls, name, bases, dct):
        """Create class from the header spec"""
        try:
            # get the type from a class attribute, could prob. have used name too
            hspec = _HEADER_SPECS[dct['_type']]
            # dynamically create the namedtuple base class
            bases = bases + (namedtuple(hspec.name + 'Tuple', hspec.fields),)
            # NOTE: using setdefault allows class definition overrides
            dct.setdefault('_fmt', hspec.fmt)
        except KeyError:    # base class
            pass
        return super(_MetaLritHeader, cls).__new__(cls, name, bases, dct)
        
    def __init__(cls, name, bases, dct):
        """Called after class declaration, register each declared type in base class dict"""
        try:                                # derived class case
            cls.header_types[dct['_type']] = cls
        except (AttributeError, KeyError):  # base class case
            cls.header_types = {}

    # class methods
    @property
    def size(cls): 
        """Header size in bytes (len if reserved for # of fields)"""
        try:
            return struct.calcsize(cls._fmt)  # works for fixed length headers
        except struct.error:
            return 0                # TODO: this was necessary for ipython pretty 
                                    # printing of variable length headers, why 
                                    # is this even called when overloaded method exists?
                                    # Should this return size _fmt with 0 or max length variable part?

## Base class to execute class construction
class LritHeader(object):
    """Base class for LRIT headers"""
    __metaclass__ = _MetaLritHeader
    __type = None
    
    def __new__(cls, buf):
        """The real constructor for this class
        
        Since the concrete derived classes inherit from namedtuple it must use
        __new__ for construction.
        """
        # obtain parser for variable length headers
        if '%' in cls._fmt:
            parser = struct.Struct(cls._fmt % (len(buf) - CommonHeader.size))
        else:
            parser = struct.Struct(cls._fmt)
        return super(LritHeader, cls).__new__(cls, *parser.unpack_from(buf))
    
    @classmethod
    def parse_header(cls, buf, hdr_type=COMMON_HTYPE):
        return cls.header_types[hdr_type](buf)

    @property
    def isvalid(self): return (self._type == self.hdr_type and 
                               (self.size == 0 or    # in case dynamically sized
                                self.size == self.hdr_len))

    @property
    def size(self): return self.__class__.size  # odd that this is necessary

# TODO: make a VariableSizedHeader abstract subclass with overloaded __new__ and size?
    
## Derived classes
class CommonHeader(LritHeader):
    _type = COMMON_HTYPE
   
MIN_HEADER_LEN = CommonHeader.size      # convenience constant

class UnknownHeader(LritHeader):
    _type = UNKNOWN_HTYPE
    # define an instance method, class method won't work w/ variable len header
    @property
    def size(self): return CommonHeader.size + len(self.data)
    @property
    def isvalid(self): return False
        
class PrimaryHeader(LritHeader):
    _type = PRIMARY_HTYPE
    
class ImageStructHeader(LritHeader):
    _type = IMAGE_STR_HTYPE
    # iscompressed field = 0 (no compression), 1 (lossless compression), 2 (lossy (JPEG) compression)
    
class ImageNavHeader(LritHeader):
    _type = IMAGE_NAV_HTYPE
    # TODO: see "CGMS 03 - LRIT/HRIT Global Spec. 2.6" sec 4.4.3.2 for projections info and 4.4 for scaling
    
class ImageDataHeader(LritHeader):
    _type = IMAGE_DAT_HTYPE
    @property
    def size(self): return CommonHeader.size + len(self.data)
    # TODO: see "CGMS 03 - LRIT/HRIT Global Spec. 2.6" sec 4.3.2 for format
    
class AnnotationHeader(LritHeader):
    _type = ANNOTATION_HTYPE
    @property
    def size(self): return CommonHeader.size + len(self.text)
    

import datetime as dt
# Tuple for CCSDS time
class CCSDSTimeTuple(namedtuple('CCSDSTimeTuple', 'pfield, days, milliseconds')):
    __slots__ = ()
    _fmt = '!BHL'
    def __new__(cls, buf):
        return super(CCSDSTimeTuple, cls).__new__(cls, *struct.unpack(cls._fmt, buf))
        
    def totimedelta(self):
        return dt.timedelta(days=self.days, milliseconds=self.milliseconds)
        
    def todatetime(self, since=dt.datetime(1958, 1, 1)):
        """Returns a datetime according to CCSDS Time Code Formats -- 3.2 CCSDS Unsegmented Time Code (CUC)"""
        # assumes time code id = default (see TimeStampHeader notes)
        return since + self.totimedelta()

class TimeStampHeader(LritHeader):
    _type = TIMESTAMP_HTYPE

     # NOTE: from CCSDS Time Code Formats -- 3.2 CCSDS Unsegmented Time Code (CUC)
     # 7 byte timestamp is (LSB):
     # 1 byte P-field:| extension flag | time code id | octets in time unit - 1 | octects in fractional unit - 1 |
     #                |    1 bit       |    3 bits    |        2 bits           |        2 bits                  |
     #    extension flag: 0 = no ext
     #    time code id: 001 = 1/1/1958, 010 = agency defined
     #  Expected P-field (LSB): b'01001011' = 0x4C
     #  Actually received: 0x40 <-- perhaps time unit octets are _in addition to_ the default
     # 6 byte T-field:|  day counter | ms of day |
     #                |    2 bytes   |  4 bytes  |
     #    day counter (2 bytes covers approx 179 years) (expected day 0 is 1/1/1958)
     #    ms of the day (1 day = 86400000 ms = 27 bits)
    def __new__(cls, buf):
        """Parse timestamp bytes into timestamp tuple"""
        obj = super(TimeStampHeader, cls).__new__(cls, buf)
        ts = CCSDSTimeTuple(obj.timestamp)
        if ts.pfield != 0x40:
            warnings.warn('Received p-field 0x%02X, expected 0x40' % ts.pfield)
        return obj._replace(timestamp=ts.todatetime())
        
class TextHeader(LritHeader):
    _type = TEXT_HTYPE
    @property
    def size(self): return CommonHeader.size + len(self.text)
        
class KeyStrHeader(LritHeader):
    _type = KEY_HTYPE
    @property
    def size(self): return CommonHeader.size + len(self.key)

class SegmentIdHeader(LritHeader):
    _type = SEG_ID_HTYPE

class NOAALRITHeader(LritHeader):
    _type = NOAA_LRIT_HTYPE
    @property
    def isvalid(self): return self.sig == "NOAA" and super(NOAALRITHeader, self).isvalid
    
class HeaderStrHeader(LritHeader):
    _type = HEADER_STR_HTYPE
    
class RiceHeader(LritHeader):
    _type = RICE_HTYPE
    @property
    def isvalid(self): return (4 <= self.pxl_per_blk <= 64 and 
                               self.pxl_per_blk % 2 == 0   and
                               self.nlines > 0             and
                               super(RiceHeader, self).isvalid)
    
# if rice header is missing, this is the default
RICE_HEADER_DEFAULT = RiceHeader(struct.pack(RiceHeader._fmt, 
                                             RiceHeader._type, 
                                             RiceHeader.size, 49, 16, 1))

class FilenameHeader(LritHeader):
    _type = FILENAME_HTYPE
    @property
    def size(self): return CommonHeader.size + len(self.text)


################################################################################
#                                                                              #
# Header structures section (see Table 15, LRIT RX Spec)                       #
#                                                                              #
################################################################################

# TODO: the header structure data field describes the subsequent headers in the 
# packet.  This could be a way of parsing subsequent headers on the fly from the
# header structure description.  This could make the code future proof if the 
# header structure sent with each transmission is accurate.
_HEADER_STR_FIELDS = (
    'NOAALRIT',        # NOAA LRIT header type
    'NLfieldlen',      # NOAA LRIT header record length
    'NLagency',        # Agency signature
    'NLprodID',        # Product ID
    'NLprodSubID',     # Product SubID
    'NLprodParm',      # Parameter
    'NLcompressflag',  # NOAA-specific compression
    'ImageStruct',     # Image Structure header type
    'ISfieldlen',      # Image Structure header record length
    'ISbitsperpix',    # Number of bits per pixel
    'ISimagecols',     # Number of columns in image
    'ISimagelines',    # Number of lines in image
    'IScompressflag',  # Compression flag
    'ImageNav',        # Image Navigation header type
    'INfieldlen',      # Image Navigation header record length
    'INprojection',    # Projection name
    'INcolscalefctr',  # Column scaling factor
    'INlinescalefctr', # Line scaling factor
    'INcoloffset',     # Column offset
    'INlineoffset',    # Line offset
    'ImageDataFunct',  # Image Data Function header type
    'IDFfieldlen',     # Image Data Function header record length
    'IDFdatadef',      # Data definition block
    'ANNOtation',      # Annotation header type
    'ANNOfieldlen',    # Annotation header record length
    'ANNOtxt',         # Annotation text
    'TimeStamp',       # Time Stamp header type
    'TSfieldlen',      # Time Stamp header record length
    'TSccsdspfield',   # CCSDS p-field
    'TSccsdsday',      # CCSDS day
    'TSccsdsms',       # CCSDS millisecond of day
    'ANCillary',       # Ancillary header type
    'ANCfieldlen',     # Ancillary header record length
    'ANCtxt',          # Ancillary text
    'SegmentID',       # Segment ID header type
    'SIDfieldlen',     # Segment ID header record length
    'SIDsequencenum',  # Image identifier
    'SIDstartcol',     # Start column of segment
    'SIDstartrow',     # Start line of segment
    'SIDtotalnum',     # Max segment
    'SIDmaxcol',       # Max column
    'SIDmaxrow',       # Max row
    'RiceID',          # Rice ID header type
    'RIDfieldlen',     # Rice ID header record length
    'RIDflags',        # Rice compression option flags
    'RIDpixel',        # Number of pixels in each CDS
    'RIDline',         # Number of compressed scan lines in one packet
)


